from typing import Dict, List, Union


class BertRouge:
    def __init__(self, tokenizer):
        self.tokenizer = tokenizer

    def __call__(
        self,
        hyps: Union[str, List[str]],
        refs: Union[str, List[str]]
    ) -> List[Dict[str, Dict[str, float]]]:
        if isinstance(hyps, str):
            hyps, refs = [hyps], [refs]

        scores: List[Dict[str, Dict[str, float]]] = []
        for hyp, ref in zip(hyps, refs):
            tokens_hyp = [token for token in self.tokenize(hyp)]
            tokens_ref = [token for token in self.tokenize(ref)]
            score: Dict[str, Dict[str, float]] = {}

            score['rouge-1'] = self.get_score(tokens_hyp, tokens_ref)
            score['rouge-2'] = self.get_score(tokens_hyp, tokens_ref, 2)

            scores.append(score)

        return scores

    def tokenize(self, text: str) -> List[str]:
        tokens: List[str] = []

        for subtoken in self.tokenizer.tokenize(text):
            if subtoken.startswith('##'):
                tokens[-1] += subtoken[2:]
            else:
                tokens.append(subtoken)

        return tokens

    def get_score(
        self,
        hyp: List[str],
        ref: List[str],
        n: int = 1
    ) -> Dict[str, float]:
        score: Dict[str, float] = {}
        ngram_hyp = {
            tuple(hyp[i:i + n])
            for i in range(len(hyp) - (n - 1))
        }
        ngram_ref = {
            tuple(ref[i:i + n])
            for i in range(len(ref) - (n - 1))
        }
        ngram_inter = ngram_hyp.intersection(ngram_ref)

        count_hyp = len(ngram_hyp)
        count_ref = len(ngram_ref)
        count_inter = len(ngram_inter)

        if count_hyp == 0:
            p = 0.0
        else:
            p = count_inter / count_hyp

        if count_ref == 0:
            r = 0.0
        else:
            r = count_inter / count_ref

        if p == 0 and r == 0:
            f = 0.0
        else:
            f = 2 * p * r / (p + r)

        score['p'] = p
        score['r'] = r
        score['f'] = f

        return score


class GreedySelector:
    def __init__(self, tokenizer, n: int = 3):
        self._rouge = BertRouge(tokenizer)
        self._n = n

    def __call__(
        self,
        sents_src: List[str],
        sents_tgt: List[str]
    ) -> list:
        reference = '\n'.join(sents_tgt)
        max_rouge = 0.0
        selected: List[int] = []
        for _ in range(self._n):
            cur_max_rouge = max_rouge
            cur_id = -1
            for i in range(len(sents_src)):
                if i in selected:
                    continue
                c = selected + [i]
                hypothesis = '\n'.join([
                    sents_src[idx]
                    for idx in c
                ])
                score = self._rouge(hypothesis, reference)
                rouge_1 = score[0]['rouge-1']['f']
                rouge_2 = score[0]['rouge-2']['f']
                rouge_score = rouge_1 + rouge_2
                if rouge_score > cur_max_rouge:
                    cur_max_rouge = rouge_score
                    cur_id = i
            if (cur_id == -1):
                break
            selected.append(cur_id)
            max_rouge = cur_max_rouge

        return [
            sents_src[i]
            for i in sorted(selected)
        ]



from copy import deepcopy
from logging import getLogger
from typing import Any, Dict, Optional

from transformers import BertConfig, BertTokenizer, EncoderDecoderConfig


logger = getLogger(__name__)


class BertSumExtConfig(BertConfig):
    def __init__(
        self,
        base_model_name_or_path: str = 'bert-base-uncased',
        encoder_num_hidden_layers: int = 2,
        encoder_num_attention_heads: int = 8,
        encoder_intermediate_size: int = 2048,
        encoder_hidden_act: str = 'gelu',
        encoder_attention_probs_dropout_prob: float = 0.1,
        encoder_layer_norm_eps: float = 1e-6,
        encoder_initializer_range: float = 0.0,
        encoder_xavier_initialization: bool = True,
        **kwargs
    ):
        config = BertConfig.from_pretrained(base_model_name_or_path) \
                           .to_dict()
        config.update(kwargs)

        encoder_config = BertConfig(
            num_hidden_layers=encoder_num_hidden_layers,
            num_attention_heads=encoder_num_attention_heads,
            intermediate_size=encoder_intermediate_size,
            hidden_act=encoder_hidden_act,
            attention_probs_dropout_prob=encoder_attention_probs_dropout_prob,
            layer_norm_eps=encoder_layer_norm_eps,
            initializer_range=encoder_initializer_range,
            xavier_initialization=encoder_xavier_initialization
        )
        config.update(encoder=encoder_config)

        super().__init__(**config)
        self.base_model_name_or_path = base_model_name_or_path

    def to_dict(self):
        output = super().to_dict()
        output['encoder'] = self.encoder.to_dict()
        return output


class BertSumAbsConfig(EncoderDecoderConfig):
    def __init__(
        self,
        encoder_model_name_or_path: str = 'bert-base-uncased',
        **kwargs
    ):
        if 'encoder' in kwargs:
            encoder_config = kwargs.pop('encoder')
        else:
            encoder_config = BertConfig.from_pretrained(encoder_model_name_or_path) \
                                       .to_dict()

        if 'decoder' in kwargs:
            decoder_config = kwargs.pop('decoder')
        else:
            decoder_config = deepcopy(encoder_config)
            decoder_config.update(kwargs)
        decoder_config['is_decoder'] = True
        decoder_config['add_cross_attention'] = True

        logger.info(f'encoder_config={encoder_config}')
        logger.info(f'decoder_config={decoder_config}')

        super().__init__(encoder=encoder_config, decoder=decoder_config)
        self.encoder_model_name_or_path = encoder_model_name_or_path



from logging import getLogger
from typing import Dict, Optional

from torch import nn
from torch.optim import AdamW
from torch.optim.lr_scheduler import LambdaLR
from tqdm import tqdm
from transformers import PreTrainedModel, Trainer, TrainingArguments

from .lr_lambda import get_transformer_schedule_with_warmup
from ..utils.nn import get_n_params

logger = getLogger(__name__)


class EncoderDecoderTrainer(Trainer):
    NO_DECAY = {'bias', 'LayerNorm.weight'}

    def __init__(
        self,
        model: PreTrainedModel,
        args: TrainingArguments,
        encoder_learning_rate: float = 0.002,
        decoder_learning_rate: float = 0.2,
        encoder_warmup_steps: int = 20000,
        decoder_warmup_steps: int = 10000,
        **kwargs
    ):
        logger.info(f'# of encoder parameters={get_n_params(model.encoder)}')
        logger.info(f'# of decoder parameters={get_n_params(model.decoder)}')

        self.is_given_optims = 'optimizers' in kwargs
        if not self.is_given_optims:
            encoder_params = self._get_params(
                model.encoder,
                args.weight_decay,
                encoder_learning_rate
            )
            decoder_params = self._get_params(
                model.decoder,
                args.weight_decay,
                decoder_learning_rate
            )
            params = encoder_params + decoder_params
            optimizer = AdamW(
                params,
                eps=args.adam_epsilon,
            )

            lr_scheduler = get_transformer_schedule_with_warmup(
                optimizer,
                [encoder_warmup_steps] * 2 + [decoder_warmup_steps] * 2
            )

            optimizers = (optimizer, lr_scheduler)
            kwargs['optimizers'] = optimizers

        super(EncoderDecoderTrainer, self).__init__(model, args, **kwargs)

    @classmethod
    def _get_params(cls, model: nn.Module, weight_decay: float, learning_rate: float) -> list:
        return [
            {
                'params': [
                    param
                    for name, param in model.named_parameters()
                    if not any(nd in name for nd in cls.NO_DECAY)
                ],
                'weight_decay': weight_decay,
                'lr': learning_rate
            },
            {
                'params': [
                    param
                    for name, param in model.named_parameters()
                    if any(nd in name for nd in cls.NO_DECAY)
                ],
                'weight_decay': 0.0,
                'lr': learning_rate
            },
        ]

    def log(self, logs: Dict[str, float]) -> None:
        lrs = self.lr_scheduler.get_last_lr()
        if self.is_given_optims:
            for i, lr in enumerate(lrs):
                logs[f'learning_rate/{i}'] = lr
        else:
            logs['learning_rate/encoder_w_decay'] = lrs[0]
            logs['learning_rate/encoder_wo_decay'] = lrs[1]
            logs['learning_rate/decoder_w_decay'] = lrs[2]
            logs['learning_rate/decoder_wo_decay'] = lrs[3]

        logs.pop('learning_rate', None)
        super().log(logs)



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

import torch
from transformers import DataCollatorWithPadding, PreTrainedTokenizer, PreTrainedTokenizerFast


@dataclass
class DataCollatorWithPaddingWithAdditionalFeatures(DataCollatorWithPadding):
    additional_features: List[str] = field(default_factory=list)

    def __call__(self, features: List[Dict[str, Union[List[int], torch.Tensor]]]) -> Dict[str, torch.Tensor]:
        additional_features = defaultdict(list)
        for feature in features:
            for key in self.additional_features:
                additional_features[key].append(feature[key])
        features = [
            {
                k: v
                for k, v in feature.items()
                if k not in self.additional_features
            }
            for feature in features
        ]

        batch = self.tokenizer.pad(
            features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors='pt',
        )

        max_length = batch['input_ids'].size(1)
        batch.update({
            key: self.pad(value, max_length)
            for key, value in additional_features.items()
        })

        if 'label' in batch:
            batch['labels'] = batch['label']
            del batch['label']

        return batch

    def pad(self, feature: List[Union[List[int], torch.Tensor]], max_length: int) -> torch.Tensor:
        matrix = []
        for row in feature:
            difference = max_length - len(row)
            row_filled = row + [0] * difference
            matrix.append(row_filled)
        return torch.tensor(matrix)


@dataclass
class EncoderDecoderDataCollatorWithPadding(DataCollatorWithPadding):
    decoder_tokenizer: Optional[Union[PreTrainedTokenizer,
                                      PreTrainedTokenizerFast]] = None
    generate_labels: bool = False

    def __post_init__(self):
        if self.decoder_tokenizer is None:
            self.decoder_tokenizer = self.tokenizer

    def __call__(self, features: List[Dict[str, Union[List[int], torch.Tensor]]]) -> Dict[str, torch.Tensor]:
        encoder_features = []
        decoder_features = []

        for feature in features:
            encoder_feature = dict()
            decoder_feature = dict()
            for k, v in feature.items():
                if k.startswith('decoder_'):
                    decoder_feature[k[len('decoder_'):]] = v
                else:
                    encoder_feature[k] = v
            encoder_features.append(encoder_feature)
            decoder_features.append(decoder_feature)

        encoder_batch = self.tokenizer.pad(
            encoder_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors='pt',
        )
        if any(decoder_features):
            decoder_batch = self.decoder_tokenizer.pad(
                decoder_features,
                padding=self.padding,
                max_length=self.max_length,
                pad_to_multiple_of=self.pad_to_multiple_of,
                return_tensors='pt',
            )
        else:
            decoder_batch = dict()

        batch = encoder_batch
        batch.update(dict(
            (f'decoder_{k}', v)
            for k, v in decoder_batch.items()
        ))

        batch['decoder_encoder_input_ids'] = batch['input_ids']

        if self.generate_labels:
            batch['labels'] = batch['decoder_input_ids']

        return batch



import re
from functools import partial, reduce
from logging import getLogger
from typing import Dict, List, Optional, Union

import spacy
from spacy.lang.en import English
from torch.utils.data import Dataset

from ...tokenizers import BertSumTokenizer, BertSumJapaneseTokenizer
from ...utils.bertsum import GreedySelector

logger = getLogger(__name__)


class BertSumDataset(Dataset):
    EMPTY_PATTERN = re.compile(r'\s+')

    def __init__(
        self,
        model_name: str,
        src: List[str],
        tgt: Optional[Union[List[str], List[List[str]]]] = None
    ):
        if tgt is not None and len(src) != len(tgt):
            raise RuntimeError('Different length src v.s. tgt pair is given.')

        # keep inputs
        self.model_name = model_name
        self.src = src
        self.tgt = tgt

        # load nlp
        if self.is_japanese:
            nlp = spacy.load('ja_ginza')
        else:
            nlp = English()
            sentencizer = nlp.create_pipe("sentencizer")
            nlp.add_pipe(sentencizer)
        self.nlp = nlp

        # create data
        encoded_src = self._encode(self.tokenizer, src)
        self.data = encoded_src

    @property
    def is_japanese(self):
        return 'bert-base-japanese' in self.model_name

    @property
    def tokenizer(self):
        if self.is_japanese:
            return BertSumJapaneseTokenizer.from_pretrained(self.model_name)
        else:
            return BertSumTokenizer.from_pretrained(self.model_name)

    def _encode(
        self,
        tokenizer: Union[BertSumTokenizer, BertSumJapaneseTokenizer],
        data: List[str],
        keep_sents: bool = True,
    ) -> List[Dict[str, List[int]]]:
        concat_sents = partial(self._concat_sents, tokenizer)
        truncate = partial(self._truncate, tokenizer)

        sentences = []
        encoded_data = []
        for text in data:
            sents = self._sentencize(text)
            if len(sents) == 0:
                continue
            if keep_sents:
                sentences.append(sents)
            n_sents = len(sents)
            sent_pairs = sum(divmod(n_sents, 2))
            outputs = []
            for i in range(sent_pairs):
                index_0 = 2 * i
                index_1 = index_0 + 1
                sent_0 = sents[index_0]
                if index_1 < n_sents:
                    sent_1 = sents[index_1]
                    output = tokenizer(sent_0, sent_1)
                else:
                    output = tokenizer(sent_0)
                outputs.append(output)

            encoded_data.append(truncate(reduce(concat_sents, outputs)))

        if keep_sents:
            self.sentences = sentences
        return encoded_data

    def _sentencize(self, text: str) -> List[str]:
        sents = [
            [str(s) for s in self.nlp(line).sents]
            for line in text.splitlines()
            if line and not self.EMPTY_PATTERN.fullmatch(line)
        ]
        if sents:
            return reduce(lambda x, y: x + y, sents)
        else:
            return []

    @staticmethod
    def _concat_sents(
        tokenizer: Union[BertSumTokenizer, BertSumJapaneseTokenizer],
        output_0: Dict[str, List[int]],
        output_1: Dict[str, List[int]]
    ) -> Dict[str, List[int]]:
        input_ids_0 = output_0['input_ids'][1:-1]
        input_ids_1 = output_1['input_ids'][1:-1]
        input_ids = tokenizer.build_inputs_with_special_tokens(input_ids_0,
                                                               input_ids_1)

        sent_sep_token_ids = tokenizer.get_sent_sep_token_ids()
        i = sum(divmod(len(sent_sep_token_ids), 2))
        token_type_ids = output_0['token_type_ids'][:-1] \
            + [1] * i \
            + [0] * (len(sent_sep_token_ids) - i) \
            + output_1['token_type_ids'][1:]

        assert len(input_ids) == len(token_type_ids), \
            'concatenated length mismatch: ' \
            f'len(input_ids)={len(input_ids)} != len(token_type_ids)={len(token_type_ids)}'

        return {
            'input_ids': input_ids,
            'token_type_ids': token_type_ids,
        }

    @staticmethod
    def _truncate(
        tokenizer: Union[BertSumTokenizer, BertSumJapaneseTokenizer],
        encoded_inputs: Dict[str, List[int]]
    ) -> Dict[str, List[int]]:
        input_ids = encoded_inputs['input_ids'][1:-1]
        num_tokens_to_remove = len(input_ids) \
            - tokenizer.model_max_length \
            + 2  # for metatoken
        input_ids, _, _ = tokenizer.truncate_sequences(
            input_ids,
            num_tokens_to_remove=num_tokens_to_remove,
            truncation_strategy='longest_first',
        )
        input_ids = tokenizer.build_inputs_with_special_tokens(input_ids)

        token_type_ids = encoded_inputs['token_type_ids'][:len(input_ids) - 1]
        token_type_ids.append(token_type_ids[-1])

        assert len(input_ids) == len(token_type_ids), \
            'truncated length mismatch: ' \
            f'len(input_ids)={len(input_ids)} != len(token_type_ids)={len(token_type_ids)}'

        return {
            'input_ids': input_ids,
            'token_type_ids': token_type_ids
        }

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx: int) -> Dict[str, List[int]]:
        return self.data[idx]


class BertSumExtDataset(BertSumDataset):
    def __init__(
        self,
        model_name: str,
        src: List[str],
        tgt: Optional[Union[List[str], List[List[str]]]] = None
    ):
        super().__init__(model_name, src, tgt)

        if self.tgt is None:
            return

        tokenizer = self.tokenizer
        bos_token_id = tokenizer.cls_token_id
        eos_token_id = tokenizer.sep_token_id
        self.gs = GreedySelector(tokenizer)

        for data in self.data:
            data['cls_mask'] = [
                1 * (id_ == bos_token_id)
                for id_ in data['input_ids']
            ]

        generate_tgt = isinstance(self.tgt[0], str)
        valid_data = []
        valid_sentences = []
        valid_tgt = []
        invalid_data = []
        for data, sents_src, sents_tgt in zip(self.data, self.sentences, self.tgt):
            if generate_tgt:
                sents_tgt = self._sentencize(sents_tgt)
                sents_tgt = self.gs(sents_src, sents_tgt)
            else:
                sents_tgt = [
                    sent
                    for text in sents_tgt
                    for sent in self._sentencize(text)
                ]  # to support multiple sentences in one sentence

            data['label'] = []
            index = 0
            for m in data['cls_mask']:
                if m:
                    sent = sents_src[index]
                    data['label'].append(
                        1 * any(st in sent for st in sents_tgt)
                    )
                    index += 1
                else:
                    data['label'].append(0)

            if sum(data['label']):
                valid_data.append(data)
                valid_sentences.append(sents_src)
                valid_tgt.append(sents_tgt)
            else:
                logger.warning(
                    'Invalid src-tgt pair was given. There are no labels.'
                )
                logger.debug(
                    f'src sentences: {sents_src}\n'
                    f'tgt sentences: {sents_tgt}'
                )
                invalid_data.append({
                    'src': sents_src,
                    'tgt': sents_tgt,
                })

        self.data = valid_data
        self.sentences = valid_sentences
        self.tgt = valid_tgt
        self.invalid_data = invalid_data


class BertSumAbsDataset(BertSumDataset):
    TGT_CLS_TOKEN = '[unused1]'
    TGT_SEP_TOKEN = '[unused2]'
    TGT_ADDITIONAL_SPECIAL_TOKENS = ['[unused3]']

    def __init__(
        self,
        model_name: str,
        src: List[str],
        tgt: Optional[List[str]] = None
    ):
        super().__init__(model_name, src, tgt)

        if tgt is None:
            return

        tokenizer = self.tgt_tokenizer
        encoded_tgt = self._encode(tokenizer, tgt, False)
        for data, e_tgt in zip(self.data, encoded_tgt):
            data.update({
                f'decoder_{k}': v
                for k, v in e_tgt.items()
            })

        # set meta objects
        vocab_size = tokenizer.vocab_size
        for token in [self.TGT_CLS_TOKEN, self.TGT_SEP_TOKEN] + self.TGT_ADDITIONAL_SPECIAL_TOKENS:
            if token not in tokenizer.vocab:
                vocab_size += 1
        self.vocab_size = vocab_size

    @property
    def tgt_tokenizer(self):
        if self.is_japanese:
            return BertSumJapaneseTokenizer.from_pretrained(
                self.model_name,
                cls_token=self.TGT_CLS_TOKEN,
                sep_token=self.TGT_SEP_TOKEN,
                additional_special_tokens=self.TGT_ADDITIONAL_SPECIAL_TOKENS
            )
        else:
            return BertSumTokenizer.from_pretrained(
                self.model_name,
                cls_token=self.TGT_CLS_TOKEN,
                sep_token=self.TGT_SEP_TOKEN,
                additional_special_tokens=self.TGT_ADDITIONAL_SPECIAL_TOKENS
            )



from typing import List, Union

from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR


class TransformerScheduler:
    def __init__(self, num_warmup_steps: int):
        self.num_warmup_steps = num_warmup_steps

    def __call__(self, current_step: int) -> float:
        if current_step < self.num_warmup_steps:
            return current_step * self.num_warmup_steps ** -1.5
        return current_step ** -.5


def get_transformer_schedule_with_warmup(
    optimizer: Optimizer,
    num_warmup_steps: Union[int, List[int]],
    last_epoch: int = -1
) -> LambdaLR:
    if isinstance(num_warmup_steps, int):
        lr_lambda = TransformerScheduler(num_warmup_steps)
    else:
        lr_lambda = [
            TransformerScheduler(n)
            for n in num_warmup_steps
        ]

    return LambdaLR(
        optimizer,
        lr_lambda,
        last_epoch=last_epoch
    )



from logging import getLogger
from typing import Dict, List, Optional

from onmt.decoders import TransformerDecoder
from onmt.modules import Embeddings
from onmt.utils.loss import LabelSmoothingLoss
import torch
from torch import nn
from torch.nn.init import xavier_uniform_
from transformers import (
    BertConfig,
    BertPreTrainedModel,
    BertModel,
    BertLMHeadModel,
    EncoderDecoderModel
)
from transformers.models.bert.modeling_bert import (
    BertEncoder,
    BertPooler,
    BertOnlyMLMHead
)
from transformers.modeling_outputs import CausalLMOutputWithCrossAttentions, SequenceClassifierOutput

from ..config import BertSumExtConfig, BertSumAbsConfig

logger = getLogger(__name__)


class BertSumExt(BertPreTrainedModel):
    config_class = BertSumExtConfig

    def __init__(self, config: BertSumExtConfig):
        super().__init__(config)

        self.bert = BertModel.from_pretrained(config.base_model_name_or_path)
        self.encoder = nn.TransformerEncoder(
            nn.TransformerEncoderLayer(
                config.hidden_size,
                config.encoder.num_attention_heads,
                dim_feedforward=config.encoder.intermediate_size,
                dropout=config.encoder.attention_probs_dropout_prob,
                activation=config.encoder.hidden_act,
            ),
            config.encoder.num_hidden_layers,
            nn.LayerNorm(config.hidden_size, eps=config.encoder.layer_norm_eps)
        )
        self.classifier = nn.Linear(config.hidden_size, 1, bias=True)
        self.loss = nn.BCEWithLogitsLoss(reduction='none')

        if config.encoder.initializer_range != 0.0:
            for p in self.encoder.layers.parameters():
                p.data.uniform_(
                    -config.encoder.initializer_range,
                    config.encoder.initializer_range
                )
        if config.encoder.xavier_initialization:
            for p in self.encoder.layers.parameters():
                if p.dim() > 1:
                    xavier_uniform_(p)

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        cls_mask=None,
        labels=None,
        return_dict=None,
        **kwargs
    ):
        outputs = self.bert(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            return_dict=return_dict,
            **kwargs
        )

        sequence_output = outputs[0].transpose(0, 1)
        cls_output = self.encoder(
            sequence_output,
            src_key_padding_mask=cls_mask.bool() ^ True,
        )
        cls_output = cls_output.transpose(0, 1)

        logits = self.classifier(cls_output).squeeze(2)

        loss = None
        if labels is not None:
            loss = self.loss(logits, labels.float())
            loss = (loss * cls_mask.float()).sum(1).mean()

        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


class BertSumAbsDecoder(BertLMHeadModel):
    # TODO: Replace with transformers decoder.
    # TODO: Control decoder and loss function arguments
    def __init__(self, config: BertConfig):
        super(BertPreTrainedModel, self).__init__(config)

        self.decoder = TransformerDecoder(
            config.num_hidden_layers,
            config.hidden_size,
            config.num_attention_heads,
            config.intermediate_size,
            False,
            'scaled-dot',
            config.hidden_dropout_prob,
            config.attention_probs_dropout_prob,
            Embeddings(
                config.hidden_size,
                config.vocab_size,
                config.pad_token_id,
                position_encoding=True
            ),
            0,
            False,
            False,
            0,
            0
        )

        self.generator = nn.Sequential(
            nn.Linear(config.hidden_size, config.vocab_size),
            nn.LogSoftmax(dim=-1)
        )

        for module in self.decoder.modules():
            if isinstance(module, (nn.Linear, nn.Embedding)):
                module.weight.data.normal_(mean=0.0, std=0.02)
            elif isinstance(module, nn.LayerNorm):
                module.bias.data.zero_()
                module.weight.data.fill_(1.0)
            if isinstance(module, nn.Linear) and module.bias is not None:
                module.bias.data.zero_()
        for p in self.generator.parameters():
            if p.dim() > 1:
                xavier_uniform_(p)
            else:
                p.data.zero_()
        self.generator[0].bias.requires_grad = False

        self.loss = LabelSmoothingLoss(
            0.1,
            config.vocab_size,
            ignore_index=config.pad_token_id
        )

        self.tie_weights()

    def get_input_embeddings(self) -> nn.Module:
        return self.decoder.embeddings.make_embedding.emb_luts[0]

    def set_input_embeddings(self, embeddings: nn.Embedding):
        self.decoder.embeddings.make_embedding.emb_luts[0] = embeddings

    def get_output_embeddings(self) -> nn.Module:
        return self.generator[0]

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        labels=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        **kwargs
    ):
        logger.debug(f'kwargs={kwargs}')

        # Use OpenNMT-py TransformerDecoder
        src = kwargs['encoder_input_ids'].unsqueeze(-1).transpose(0, 1)
        tgt = input_ids.unsqueeze(-1).transpose(0, 1)
        memory_bank = enc_hidden = encoder_hidden_states.transpose(0, 1)

        self.decoder.init_state(src, memory_bank, enc_hidden)
        dec_outs, attns = self.decoder(
            tgt,
            memory_bank,
            memory_lengths=encoder_attention_mask.sum(axis=1)
        )

        # transformers style loss calculation
        decoder_outputs = dec_outs.transpose(0, 1)
        prediction_scores = self.generator[0](decoder_outputs)

        lm_loss = None
        if labels is not None:
            # we are doing next-token prediction; shift prediction scores and input ids by one
            shifted_prediction_scores = prediction_scores[:, :-1, :] \
                .contiguous()
            labels = labels[:, 1:].contiguous()

            output = self.generator[1](
                shifted_prediction_scores
            ).view(-1, self.config.vocab_size)
            target = labels.view(-1)

            normalization = target.ne(self.config.pad_token_id).sum().item()

            lm_loss = self.loss(output, target).div(float(normalization))

        if not return_dict:
            output = (prediction_scores, None, None, attns['std'], None)
            return ((lm_loss,) + output) if lm_loss is not None else output

        return CausalLMOutputWithCrossAttentions(
            loss=lm_loss,
            logits=prediction_scores,
            past_key_values=None,
            hidden_states=None,
            attentions=None,
            cross_attentions=attns['std'],
        )


class BertSumAbs(EncoderDecoderModel):
    config_class = BertSumAbsConfig

    def __init__(
        self,
        config: Optional[BertSumAbsConfig] = None,
        encoder: Optional[BertPreTrainedModel] = None,
        decoder: Optional[BertPreTrainedModel] = None
    ):
        if config is not None:
            if encoder is None:
                encoder = BertModel.from_pretrained(
                    config.encoder_model_name_or_path
                )
            if decoder is None:
                decoder = BertSumAbsDecoder(config.decoder)

        super().__init__(config=config, encoder=encoder, decoder=decoder)

        logger.debug(f'self.config={self.config}')

        enc_emb_weight = self.encoder.get_input_embeddings().weight.clone().detach()
        dec_emb_weight = self.decoder.get_input_embeddings().weight.clone().detach()
        enc_row, enc_col = enc_emb_weight.size()
        dec_emb_weight[:enc_row, :enc_col] = enc_emb_weight
        input_embeddings = nn.Embedding.from_pretrained(
            dec_emb_weight,
            freeze=False,
            padding_idx=self.config.decoder.pad_token_id
        )
        self.decoder.set_input_embeddings(input_embeddings)
        self.decoder.tie_weights()

    def prepare_inputs_for_generation(self, input_ids, past=None, attention_mask=None, encoder_outputs=None, **kwargs):
        input_dict = super().prepare_inputs_for_generation(
            input_ids,
            past,
            attention_mask,
            encoder_outputs,
            **kwargs
        )

        if 'decoder_encoder_input_ids' in kwargs:
            input_dict['decoder_encoder_input_ids'] = kwargs['decoder_encoder_input_ids']

        return input_dict



import re
from typing import List, Optional

from transformers import BertTokenizer, BertJapaneseTokenizer


class BertSumMixin:
    def build_inputs_with_special_tokens(
        self,
        token_ids_0: List[int],
        token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        cls = [self.cls_token_id]
        sep = [self.sep_token_id]

        if token_ids_1 is None:
            return cls + token_ids_0 + sep

        sent_sep_token_ids = self.get_sent_sep_token_ids()

        return cls \
            + token_ids_0 \
            + sent_sep_token_ids \
            + token_ids_1 \
            + sep

    def create_token_type_ids_from_sequences(
        self,
        token_ids_0: List[int],
        token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        cls = [self.cls_token_id]
        sep = [self.sep_token_id]

        if token_ids_1 is None:
            return len(cls + token_ids_0 + sep) * [0]

        sent_sep_token_ids = self.get_sent_sep_token_ids()
        i = sum(divmod(len(sent_sep_token_ids), 2))
        former_sent_sep_token_ids = sent_sep_token_ids[:i]
        latter_sent_sep_token_ids = sent_sep_token_ids[i:]

        return len(cls + token_ids_0 + former_sent_sep_token_ids) * [0] \
            + len(latter_sent_sep_token_ids + token_ids_1 + sep) * [1]

    def get_sent_sep_token_ids(self) -> List[int]:
        if self.additional_special_tokens_ids:
            sent_sep_token_ids = self.additional_special_tokens_ids
        else:
            sent_sep_token_ids = [self.sep_token_id, self.cls_token_id]

        return sent_sep_token_ids


class BertSumTokenizer(BertSumMixin, BertTokenizer):
    pass


class BertSumJapaneseTokenizer(BertSumMixin, BertJapaneseTokenizer):
    ASCII_PATTERN = re.compile(r'[A-Za-z]+')

    @classmethod
    def clean_up_tokenization(cls, out_string: str) -> str:
        tokens = out_string.split()
        jp_tokens = []
        sub_tokens = []
        for token in tokens:
            if cls.ASCII_PATTERN.fullmatch(token):
                sub_tokens.append(token)
            else:
                if sub_tokens:
                    jp_tokens.append(' '.join(sub_tokens))
                    sub_tokens = []
                jp_tokens.append(token)
        out_string = ''.join(jp_tokens)
        out_string = BertJapaneseTokenizer.clean_up_tokenization(out_string)
        return out_string
